const thunky = require('thunky')

const hypercore = require('hypercore')
const MountableHypertrie = require('mountable-hypertrie')

const { contentKeyPair, contentOptions } = require('./lib/content')

class HyperdriveState {
  constructor (factory, key, opts) {
    this.key = key

    this.live = true
    this.latest = !!opts.latest
    this.sparse = !!opts.sparse
    this.sparseMetadata = !!opts.sparseMetadata
    this._factory = opts.factory

    this._db = opts._db || new MountableHypertrie(this._factory, key, {
      secretKey: opts.secretKey,
      sparse: this.sparseMetadata,
      createIfMissing: opts.createIfMissing,
      storageCacheSize: opts.metadataStorageCacheSize,
      valueEncoding: 'binary'
    })
    this.content = opts.content || null

    this._db = null
    this._contentOpts = null
    this._contentFeedLength = null
    this._contentFeedByteLength = null
    this._contentKeyPair = null
    
    this.contentReady = thunky(this._contentReady.bind(this))
    this.ready = thunky(this._ready.bind(this))
  }

  _createHypercore (storage, key, opts) {
    if (this._factory) return this._factory(key, opts)
    return hypercore(storage, key, opts)
  }

  _contentReady (cb) {
    this.ready(err => {
      if (err) return cb(err)
      if (this.content) return cb(null)
      this._ensureContent(null, cb)
    })
  }

  _ready () {
    const self = this

    this.metadata.on('error', onerror)
    this.metadata.on('append', update)

    this.metadata.ready(err => {
      if (err) return cb(err)

      if (this.sparseMetadata) {
        this.metadata.update(function loop () {
          self.metadata.update(loop)
        })
      }

      this._contentKeyPair = this.metadata.secretKey ? contentKeyPair(this.metadata.secretKey) : {}
      this._contentOpts = contentOptions(this, this._contentKeyPair.secretKey)
      this._contentOpts.keyPair = this._contentKeyPair

      /**
       * If a db is provided as input, ensure that a contentFeed is also provided, then return (this is a checkout).
       * If the metadata feed is writable:
       *    If the metadata feed has length 0, then the db should be initialized with the content feed key as metadata.
       *    Else, initialize the db without metadata and load the content feed key from the header.
       * If the metadata feed is readable:
       *    Initialize the db without metadata and load the content feed key from the header.
       */
      if (this._db) {
        if (!this.content || !this.metadata) return cb(new Error('Must provide a db and both content/metadata feeds'))
        return done(null)
      } else if (this.metadata.writable && !this.metadata.length) {
        initialize(this._contentKeyPair)
      } else {
        restore(this._contentKeyPair)
      }
    })

    /**
     * The first time the hyperdrive is created, we initialize both the db (metadata feed) and the content feed here.
     */
    function initialize (keyPair) {
      self._ensureContent(keyPair.publicKey, err => {
        if (err) return cb(err)
        self._db = hypertrie(null, {
          feed: self.metadata,
          metadata: self.content.key
        })

        self._db.ready(function (err) {
          if (err) return cb(err)
          return done(null)
        })
      })
    }

    /**
     * If the hyperdrive has already been created, wait for the db (metadata feed) to load.
     * If the metadata feed is writable, we can immediately load the content feed from its private key.
     * (Otherwise, we need to read the feed's metadata block first)
     */
    function restore (keyPair) {
      self._db = hypertrie(null, {
        feed: self.metadata
      })
      if (self.metadata.writable) {
        self._db.ready(err => {
          if (err) return done(err)
          self._ensureContent(null, done)
        })
      } else {
        self._db.ready(done)
      }
    }

    function done (err) {
      if (err) return cb(err)
      self.key = self.metadata.key
      self.discoveryKey = self.metadata.discoveryKey
      return cb(null)
    }

    function onerror (err) {
      if (err) self.emit('error', err)
    }

    function update () {
      self.emit('update')
    }
  }
}

module.exports = HyperdriveState
